#!/usr/bin/perl

#------------------------------------------------------------------------------
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#------------------------------------------------------------------------------


# Allow mounting of Acorn DFS disks and disk image files
# using Fuse (http://fuse.sourceforge.net/)
#
# $Header: /var/cvs/Fuse-DFS/Fuse-DFS.pl,v 1.2 2006/12/25 16:45:33 adsb Exp $
# Andrew Benham (adsb@adsb.co.uk) 2006

use warnings;
use strict;
use Fuse;
use threads;
use threads::shared;
use POSIX qw(ENOENT EISDIR EINVAL);
use File::Basename;
use Fcntl ':mode';
use Getopt::Std;
use Data::Dumper;

# Default settings DFS.
my $sector_size = 256;          # 256 bytes/sector
my $entries_per_dir = 31;       # 31 file system objects per directory
my $file_size_in_dir = 8;       # 8 octets per entry in directory
my $sectors_per_dir = 2;        # 2 sectors in directory

# Get the first group number from the $( list (current user's groups)
my $primary_group = ( split /\s+/, $( )[0];     #)  Keep bracket matchers happy

my %opts;
getopts('dl2S', \%opts);

my $debug = $opts{d} ? 1 : 0;
my $print_listing = $opts{l} ? 1 : 0;
my $second_catalog = $opts{2} ? 1 : 0;
my $solidisk_ddfs = $opts{S} ? 1 : 0;

$sectors_per_dir = 4 if $second_catalog;

unless (scalar @ARGV == 2) {
        warn "Syntax: $0 [-dl2S] <device> <mountpoint>\n";
        warn "\t-d\tdebug output on\n"; 
        warn "\t-l\tlisting output on\n"; 
        warn "\t-2\tuse second catalog\n"; 
        warn "\t-S\tdisk is Solidisk DDFS\n"; 
        exit 1;
}
my $device = shift(@ARGV);
my $mountpoint = shift(@ARGV);

my %diskinfo = (
        size => undef,
        opt4 => undef,
        name => undef,
        msn => undef,
        freespace => undef,
);                

open FILE,$device or die "Can't open $device; $!\n";

my $directory;
if ($second_catalog) {
        seek FILE, 512, 0  or die "Can't seek to second catalog\n";
}
die "Can't read directory\n" unless read(FILE,$directory,512) == 512;

# We store the file details in a hash, we need to start
# with a top-level directory.  We also add a '..THIS_DISK.inf'
# file to make visible the attributes of the DFS disk.
my %files = (
        '/' => {
                type => S_IFDIR,
                mode => 0755,
                ctime => 0,
                length => 0,
        },
);

die "Can't understand directory - not valid DFS ?\n"
        unless parse_directory($directory,\%diskinfo);

$files{'/..THIS_DISK.inf'} = {
                type => S_IFREG,
                mode => 0444,
                ctime => 0,
                content => "*OPT4,$diskinfo{opt4}\n" .
                            "T: $diskinfo{name}, S: $diskinfo{msn}\n",
                };

# Run the Fuse main loop.  This doesn't return until you 'fusermount -u'
# the mount point.
my $mnt = Fuse::main(
        #debug => 1,
        mountpoint => $mountpoint,
        mountopts => 'allow_other,fsname=dfs',
        threaded => 1,
        getattr => 'main::dfs_getattr',
        getdir => 'main::dfs_getdir',
        read => 'main::dfs_read',
        statfs => 'main::dfs_statfs',
);

close FILE;


# Walk a DFS directory
sub parse_directory {
        my $directory = shift;
        my $diskinfo = shift;

        my $sector0 = substr($directory, 0, $sector_size);
        my $sector1 = substr($directory, $sector_size, $sector_size);

        my $title = unpack("a8", $sector0);
        my ($more_title, $msn, $num_entries, $options, $num_secs) =
            unpack("a4 C C C C", $sector1);
        $title .= $more_title;
        $diskinfo->{name} = $title;

        $diskinfo->{msn} = $msn;

        $diskinfo->{opt4} = ($options & 0x30) >> 4;

        # The number of sectors on the floppy is a 10-bit number
        # in Acorn's docs, which allows disks up to 1024 sectors.
        # Watford and Solidisk produced double-density versions
        # of DFS, using 16 sectors per track - so they use 11-bit
        # numbers - stretching into an unused bit in sector 1 byte 6.
        my $extra_num_secs = ($options & 0x07);
        $num_secs += ($extra_num_secs << 8);
        $diskinfo->{size} = $num_secs;

        my $file_count = 0;
        my $used_sector_count = $sectors_per_dir;       # Directory sectors

        foreach my $i (0 .. $entries_per_dir - 1) {
                last if $i *8  == $num_entries;
                $file_count++;
                my $file_name = substr($sector0, ($i * $file_size_in_dir) + 8,
                                       $file_size_in_dir);
                my $file_entry = substr($sector1, ($i * $file_size_in_dir) + 8,
                                        $file_size_in_dir);
                my ($load, $exec, $length, $extra, $start) =
                    unpack("v v v C C",$file_entry);

                my $extra_exec   = ($extra & 0xc0) >> 6;
                my $extra_length = ($extra & 0x30) >> 4;
                my $extra_load   = ($extra & 0x0c) >> 2;
                my $extra_start  = ($extra & 0x03);

                # Solidisk's DDFS needs an extra bit for each of
                # start sector and file length because the disk
                # size is larger than that of Acorn's.  They use
                # the extra two bit of the load address in a completely
                # different way from Acorn - so we only cope with this
                # perversion if explicitly told this is a Solidisk DDFS
                # disk, otherwise we're prone to false matching.

                if ($solidisk_ddfs) {
                        if ($extra_load & 1) {
                                $extra_start += (1 << 2);
                        } elsif ($extra_load & 2) {
                                $extra_length += (1 << 2);
                        }

                        # Now, what do we do as regards the 'real'
                        # extra load address bits ?
                        # It's not actually possible to have both
                        # the 'extra length bit' and the 'extra start
                        # sector' bit set together (the disk isn't large
                        # enough for that).  
                        # So are '11' and '00' used as the only valid values
                        # for the extra load address bits ?
                        # Or should we copy the extra exec address bits
                        # into the extra load address bits ?
                        # I've mislaid my Solidisk manual, so I'm
                        # guessing here.
                        $extra_load  = $extra_exec;
                }

                $extra_exec = 0xffff if $extra_exec == 3;
                $extra_load = 0xffff if $extra_load == 3;

                $exec   += ($extra_exec   << 16);
                $length += ($extra_length << 16);
                $load   += ($extra_load   << 16);
                $start  += ($extra_start  << 8);

                $used_sector_count += POSIX::ceil($length/$sector_size);

                my @name = unpack("C8",$file_name);
                my $ctime = 0;
                last unless $name[0];   # Detect end of entries
                my $leafname = '';
                my $finished_name = 0;
                for my $j (0 .. 6) {
                        my $c = $name[$j];
                        my $c1 = $c & 0x7f;
                        $finished_name = 1 if $c1 == ord ' ';
                        $leafname .= chr $c1 unless $finished_name;
                }
                my $directory_byte = $name[7] & 0x7f;
                my $locked_bit = $name[7] >> 7;

                # Watford DDFS need extra bits for the length and
                # start sector fields to cope with their disks being
                # slightly larger than Acorn's.  They store these
                # extra bits as the top bits of two of the filename
                # chars, which are normally clear so it's safe to do
                # this on any disk.
                my $watford_extra_length = $name[5] >> 7;
                my $watford_extra_start  = $name[6] >> 7;
                $length += ($watford_extra_length << 18);
                $start  += ($watford_extra_start  << 10);

                if ($directory_byte != ord '$') {
                        $directory_byte = chr $directory_byte;
                        $leafname = $directory_byte . '.' . $leafname;
                }
                my $realname = '/' . $leafname;
                my $infname = '/.' . $leafname . '.inf';
                printf("%-11s: %08lx %08lx %08lx %6x %s\n",
                        $realname, $load, $exec, $length,
                        $start, $locked_bit ? 'L' : '')
                        if $print_listing;
                my $mode = $locked_bit ? 0555: 0777;

                # If the special 'other catalog' file is present
                # then this disk has two catalogs.  We shouldn't
                # display this file.
                if ($leafname eq '!.!!!!!!!' and $locked_bit) {
                        # Record the fact that we have two catalogs
                        # unless we aleady know.
                        unless ($sectors_per_dir == 4) {
                                $sectors_per_dir = 4;
                                $used_sector_count += 2;
                        }
                } else {
                        $files{$realname} = {
                                type => S_IFREG,
                                mode => $mode,
                                ctime => $ctime,
                                length => $length,
                                sector => $start,
                        };
                        $files{$infname} = {
                                type => S_IFREG,
                                mode => $mode,
                                ctime => 0,
                                content => sprintf("$leafname, " .
                                           "L: %06X, E: %06X F:%s\n",
                                           $load, $exec,
                                           $locked_bit ? 'L' : ''),
                        };
                }
        }
        $diskinfo->{filecount} = $file_count;
        $diskinfo->{freespace} = $diskinfo->{size} - $used_sector_count;
        return 1;
}

sub dfs_getattr {
        my $file = shift;
        print "getattr: $file\n" if $debug;
        return -ENOENT() unless exists($files{$file});
        my ($nlink, $blksize) = (1, $sector_size);
        my ($size, $blocks);
        if (exists $files{$file}{length}) {
                $size = $files{$file}{length};
                $size = 0 if $files{$file}{type} == S_IFDIR;
                $blocks = POSIX::ceil($size/$blksize);
        } elsif (exists $files{$file}{content}) {
                # Not a real DFS file
                $size = length $files{$file}{content};
                $blocks = 0;
        } else {
                die "Can't get size of '$file'\n";
        }
        my $modes = $files{$file}{type} + $files{$file}{mode};
        my ($dev, $ino, $rdev, $uid, $gid) =
           (0,    0,    0,     $<,   $primary_group);
        if (exists $files{$file}{nlinks}) {
                $nlink = $files{$file}{nlinks};
        }
        my ($atime, $ctime, $mtime);
        $atime = $ctime = $mtime = $files{$file}{ctime};
        return ($dev, $ino, $modes, $nlink, $uid, $gid, $rdev, $size,
                $atime, $mtime, $ctime, $blksize, $blocks);
}

sub dfs_getdir {
        my $dir = shift;
        print "getdir: $dir\n" if $debug;
        $dir .= '/' unless $dir eq '/';
        my @files = grep {m,^$dir[^/]*$,} (sort keys %files);
        @files = map { basename $_ } @files;
        @files = grep { ! m{^/$} } @files;
        unshift @files, '.','..';
        push @files, 0;
        return @files;
}

sub dfs_read {
        my $pathname = shift;
        my $size = shift;
        my $offset = shift;
        
        print "read: $pathname, $size bytes starting at $offset\n" if $debug;
        return -ENOENT() unless exists($files{$pathname});

        my $buffer;
        if (exists $files{$pathname}{sector}) {
                my $start_sector = $files{$pathname}{sector};
                print "   starts at sector $start_sector\n" if $debug;
                my $length = $files{$pathname}{length};
                print "   length = $length\n" if $debug;
                $length = $size if $length > $size;
                read_starting_from_sector(\$buffer, $start_sector, $length,
                                          $offset);
        } elsif (exists $files{$pathname}{content} ) {
                $buffer = substr $files{$pathname}{content}, $offset, $size;
        } else {
                die "Don't know how to read '$pathname'\n";
        }
        return $buffer;
}

sub dfs_statfs {
        # Filename length, total inodes, free inodes, total blocks,
        # free blocks, block size
        # Total blocks is disk size less the directory.

        my $available_sectors = $diskinfo{size} - $sectors_per_dir;
        my $free_sector_count = $diskinfo{freespace};

        my $available_files = $entries_per_dir;
        my $free_files = $entries_per_dir - $diskinfo{filecount};

        print "statfs: available sectors = $available_sectors, ",
              "free = $free_sector_count\n" if $debug;
        print "        available files = $available_files, ",
              "free = $free_files\n" if $debug;

        return 9, $available_files, $free_files,
               $available_sectors, $free_sector_count, $sector_size;
}

# -----------------------------------------------------------------------------
# Read $length bytes starting at DFS sector $start_sector (with an optional
# offset of $offset bytes) into $buffer

sub read_starting_from_sector {
        my $buffer = shift;
        my $start_sector = shift;
        my $length = shift;
        my $offset = shift || 0;

        my $start_pos = $offset + ($start_sector * $sector_size);

        my $bytes_read;
        seek FILE, $start_pos, 0;
        $bytes_read = read FILE, $$buffer, $length;

        return $bytes_read;
}
